#include "xmodem.h"

#include <termios.h>

#include <signal.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <unistd.h>

#include <stdio.h>

/* Line control codes */
#define SOH			0x01	/* start of header */
#define STX			0x02	/* start of header, 1k block size */
#define ACK			0x06	/* Acknowledge */
#define NAK			0x15	/* Negative acknowledge */
#define CAN			0x18	/* Cancel */
#define EOT			0x04	/* end of text */

char xmodem_buffer[XMODEM_BLOCK_SIZE];

static int fd;

static int blocknum;

static void put_byte(unsigned int c);
static int get_byte(void);

/* #define USE_STDIN */

/* #define TESTING */

/* #define DEBUG_CHAR(x) (write(2, x, 1)) */
#define DEBUG_CHAR(x)


static struct termios save_termios, new_termios;

int xmodem_open() 
{
  char *msg = XMODEM_MSG_BEGIN;

  blocknum = 1;

#ifdef USE_STDIN
  fd = 0;
#else
  fd = open(XMODEM_TTY, O_RDWR | O_NOCTTY);
  if (fd == -1) {
    perror("open tty");
    exit(1);
  }
#endif

  if (tcgetattr(fd, &save_termios)) {
    perror("tcgetattr");
    exit(1);
  }

  if (tcgetattr(fd, &new_termios)) {
    perror("tcgetattr");
    exit(1);
  }
  
  cfmakeraw(&new_termios);
  cfsetispeed(&new_termios, B115200);
  cfsetospeed(&new_termios, B115200);

  new_termios.c_cflag |= CLOCAL;
  new_termios.c_cflag &= ~CRTSCTS;

  new_termios.c_cc[VMIN] = 0;
  new_termios.c_cc[VTIME] = 30;

  if (tcsetattr(fd, TCSANOW, &new_termios)) {
    perror("tcsetattr");
    exit(1);
  }

  sleep(1);

  fprintf(stderr, msg);
  write(fd, msg, strlen(msg));

  sleep(1);

  return 0;
}

void xmodem_restore()
{
  if (tcsetattr(fd, TCSAFLUSH, &save_termios)) {
    perror("tcsetattr");
    exit(1);
  }

#ifndef USE_STDIN
  close(fd);
#endif
}

static int read_raw_block()
{
  int total = 0;
  int count;

#if 0
  new_termios.c_cc[VMIN] = 128;
  new_termios.c_cc[VTIME] = 30;

  if (tcsetattr(fd, TCSANOW, &new_termios)) {
    perror("tcsetattr");
    exit(1);
  }
#endif

  while (total < XMODEM_BLOCK_SIZE) {
    count = read(fd, xmodem_buffer + total, XMODEM_BLOCK_SIZE - total);
    if (count == 0) {
      /* timed out */
      return 1;
    }
    if (count < 0) {
      /* printf("lost badly, %d\n", count); */
      return -1;
    } 
    total = total + count;
  }

  return 0;
}

static unsigned int matches_checksum_block(unsigned int checksum) 
{
  unsigned int c = 0;
  int i;

  for (i = 0; i < XMODEM_BLOCK_SIZE; i++) {
    c += (xmodem_buffer[i] & 0xff);
  }

  return checksum == (c & 0xff);
}

static unsigned char input_holder;

static int get_byte(void) 
{
  int n;

#if 0
  new_termios.c_cc[VMIN] = 0;
  new_termios.c_cc[VTIME] = 30;

  if (tcsetattr(fd, TCSANOW, &new_termios)) {
    perror("tcsetattr");
    exit(1);
  }
#endif

  DEBUG_CHAR("O");

  n = read(fd, &input_holder, 1);

  if (n < 0) {
    perror("xmodem header read");
    exit(1);
  }

  if (n == 0) {
    /* printf("timeout\n"); */
    DEBUG_CHAR("T");
    return -1; /* timed out */
  }

  if (n == 1) {
    DEBUG_CHAR("U");
    return input_holder & 0xff;
  }


  fprintf(stderr, "weird return from read\n");
  exit(1);
}

static char put_buffer;

static void put_byte(unsigned int c)
{
  put_buffer = c;

  if (write(fd, &put_buffer, 1) != 1) {
    perror("xmodem write");
    exit(1);
  }
  // fsync(fd);
}

static char *twirler = "/-\\|";
static char twirl_template[] = ".\010"; /* character, backspace */

int xmodem_get_record(void)
{
  unsigned char c;

  int failed;

  DEBUG_CHAR("J");
 retry:

  failed = 0;

  c = get_byte();
  DEBUG_CHAR("K");
  switch (c) {
#ifdef XMODEM_USE_1K
  case STX: /* there's a 1k packet coming */
#else
  case SOH:    /* there's a 128 byte packet coming */
#endif
    break;
  case EOT:
    put_byte(ACK);
    return 1;
    break;
  case CAN:
    put_byte(ACK);
    return -1;
    break;
  case -1:     /* timed out */
  default:
    put_byte(NAK);
    goto retry;
    break;
  }


  c = get_byte();

  /* printf("got alleged block %d\n", c); */

  if ((c & 0xff) != (blocknum & 0xff)) {
    fprintf(stderr, "failing on blocknum %d != %d, will retry\n", c & 0xff, blocknum & 0xff);
    failed = 1;
  }

  c = get_byte();
  /* this should be the block number checksum, but who cares */
  /* if (c != (- block_num - 1) return -2; */

  c = read_raw_block();
  if (c == 1) { /* timeout */
    sleep(2);
    tcflush(fd, TCIFLUSH);
    put_byte(NAK);
    goto retry;
  }
  
  c = get_byte(); /* checksum */

  if (!matches_checksum_block(c)) {
    fprintf(stderr, "checksum failed, retrying\n");
    failed = 1;
  }

  if (failed) {
    put_byte(NAK);
    goto retry;
  }
  
  put_byte(ACK);

  twirl_template[0] = twirler[blocknum % 4];
  fputs(twirl_template, stderr);

  blocknum++;
  return 0;
}

#ifdef TESTING

int main(int argc, char **argv)
{
  int retval, i;

  xmodem_open();

  while (!(retval = xmodem_get_record())) {
    write(1, xmodem_buffer, XMODEM_BLOCK_SIZE);
  }

  if (retval == 1) {
    printf("ok, that's fine\n");
    xmodem_restore();
    exit(0);
  } else {
    printf("failed with code %d\n", retval);
    xmodem_restore();
    exit(1);
  }
}
#endif
    
/* now, the higher level stuff */

static int xmodem_valid = 0; /* anything valid in buffer? */
static int xmodem_internal_offset;  /* how far we are inside the buffer */

static int xmodem_seen_eof = 0;

int xmodem_position = 0; 

/* read at most count bytes.  return number read.  null pointer means seek */
static int xmodem_read_upto(void *ptr, int count)
{
  int remaining;

  if (!xmodem_valid) {
    int res = xmodem_get_record();
    xmodem_valid = 1;
    xmodem_internal_offset = 0;
    if (res == 1) { fprintf(stderr, "xmodem eof\n"); xmodem_seen_eof = 1; }
    if (res < 0) { fprintf(stderr, "error in xmodem, bailing out\n"); exit(1); }
  }

  remaining = XMODEM_BLOCK_SIZE - xmodem_internal_offset;

  if (count > remaining)
    count = remaining;

  if (ptr)
    memcpy(ptr, xmodem_buffer + xmodem_internal_offset, count);

  xmodem_internal_offset += count;
  xmodem_position += count;

  if (xmodem_internal_offset == XMODEM_BLOCK_SIZE)
    xmodem_valid = 0;

  return count;
}

int xmodem_read(void *ptr, int count)
{
  int n;
  int total = 0;

  while (count > 0) {
    n = xmodem_read_upto(ptr + total, count);
    count = count - n;
    total = total + n;
  }

  return 0;
}

int xmodem_seek(int pos) 
{
  int offset = pos - xmodem_position;

  if (offset == 0)
    return 0;

  /* support a limited amount of backtracking, as long as we're still
     in the same block.  good enough for image loads. */

  if (offset < 0) {
    if (pos < XMODEM_BLOCK_SIZE && xmodem_position < XMODEM_BLOCK_SIZE) {
      xmodem_internal_offset = pos;
      xmodem_position = pos;
      return 0;
    }
    fprintf(stderr, "can't seek backward to previous blocks on xmodem stream!\n");
    exit(1);
  }

  while (offset > 0) {
    offset = offset - xmodem_read_upto(0, offset);
  }

  return 0;
}

void xmodem_close(void)
{
  if (!xmodem_seen_eof) {
    while (xmodem_get_record() == 0)
      /* wait for eof */;
  }
  xmodem_restore();
}
